/****************************************************************************
** Demonstration of scrolling a tiled background.                          **
**  by Steven H Don                                                        **
**                                                                         **
** Please note: SpriteRec is the same type as used in the sprite routines  **
** on my homepage. The MakeSprite routine and some, less significant       **
** others also originate from those files.                                 **
**                                                                         **
** For questions, feel free to e-mail me.                                  **
**                                                                         **
**    shd@earthling.net                                                    **
**    http://shd.cjb.net                                                   **
**                                                                         **
****************************************************************************/
#include "malloc.h"
#include "dos.h"
#include "mem.h"
#ifdef __DJGPP
  #include "sys/nearptr.h"
#endif

struct SpriteRec {
  //These indicate the dimensions of the sprite
  unsigned int Width, Height, Total;
  //These indicate the location of the sprite in system memory
  #ifdef __DJGPP
    unsigned char *SpriteData;
  #else
    unsigned char far *SpriteData;
  #endif

  //NOTE: The following values are optional

  //These indicate the location of the sprite on screen
  unsigned int X, Y;
  //These indicate the direction of the sprite movement
  char Xdir, Ydir;
};

  //Reserve space for 16 tiles
  struct SpriteRec          Tiles [16];
  //The map of tiles itself
  char               TileMap [128*128];

  #ifdef __DJGPP
    unsigned char Buffer [64000];
    unsigned char *Screen;
  #else
    unsigned char far *Buffer;
    unsigned char far *Screen;
  #endif

//Switches to VGA 320x200 256 colour
void VGAScreen ()
{
  union REGS r;
  r.h.ah = 0;
  r.h.al = 0x13;
  int86(0x10, &r, &r);
}

//Switches to text 80x25
void TextScreen ()
{
  union REGS r;
  r.h.ah = 0;
  r.h.al = 0x3;
  int86(0x10, &r, &r);
}

//Sets a pixel in the buffer to a specific colour
void PutPixel (unsigned int x, unsigned int y, unsigned char c)
{
  if ((x < 320) && (y < 200)) {
    Buffer [(y << 8) + (y << 6) + x] = c;
  }
}

//Reads the colour value of a pixel from the buffer
unsigned char ReadPixel (unsigned int x, unsigned int y)
{
  if ((x < 320) && (y < 200)) {
    return (Buffer [(y << 8) + (y << 6) + x]);
  } else {
    return (0);
  }
}

//Clears the buffer
void ClearBuffer ()
{
  #ifdef __DJGPP
    memset (Buffer, 0, 64000);
  #else
    _fmemset (Buffer, 0, 64000);
  #endif
}

//Displays the buffer
void DisplayBuffer ()
{
  #ifdef __DJGPP
    memcpy (Screen, Buffer, 64000);
  #else
    _fmemcpy (Screen, Buffer, 64000);
  #endif
}

unsigned char MakeSprite (struct SpriteRec *Sprite,
                          unsigned int x1, unsigned int y1,
                          unsigned int x2, unsigned int y2)

{
  unsigned int      Size, x, y;
  unsigned int      CurrentPixel;
  #ifdef __DJGPP
    unsigned char *SpriteMem;
  #else
    unsigned char far *SpriteMem;
  #endif

  //Calculate amount of memory needed for sprite
  Size = (y2 - y1 + 1) * (x2 - x1 + 1) - 1;
  //Check if enough memory is available
  #ifdef __DJGPP
    if ((SpriteMem = (char *) malloc(Size))==NULL) {
  #else
    if ((SpriteMem = (char far *) farmalloc(Size))==NULL) {
  #endif
    //If not, return 0
    return (0);
  }
  Sprite->SpriteData = SpriteMem;

  //Set dimensions
  Sprite->Width   = x2 - x1;
  Sprite->Height  = y2 - y1;
  Sprite->Total   = Size + 1;
  //Increase width/height for correct loop
  x2++; y2++;
  //Read sprite from the screen buffer
  CurrentPixel = 0;
  for (y = y1; y < y2; y++) {
    for (x = x1; x < x2; x++) {
      Sprite->SpriteData [CurrentPixel++] =
        ReadPixel (x, y);
    }
  }
  //Return true
  return (1);
}

void DrawTiledBackground (unsigned int tlX, unsigned int tlY)
{
  //Temporary x and y variables
  unsigned int tX, tY;
  //Offset into the buffer, there's no need to use Putpixel when filling
  // linearly.
  unsigned int  o = 0;
  //These give the current coordinates in the tile map.
  unsigned int TMapX, TMapY;
  //These give the coordinates of a tile in the tile map.
  unsigned int TileX, TileY;
  //These give the coordinates of a pixel in the tile.
  unsigned int TPixX, TPixY;

  for (tY = 0; tY < 200; tY++) for (tX = 0; tX < 320; tX++) {
    //Calculate which pixel in the tile map is needed.
    TMapX = tlX + tX; TMapY = tlY + tY;
    /*
      Note that the following calculations assume tiles of 16x16 pixels each.
      If you change the tile size, you'll need to change the references
      to '>> 4' and '& 15' to other values.
    */
    //From that, calculate from what tile that pixel should come.
    TileX = TMapX >> 4; TileY = TMapY >> 4;
    //And which pixel in the tile should be drawn.
    TPixX = TMapX & 15; TPixY = TMapY & 15;
    //Draw the pixel
    Buffer [o++] = Tiles [TileMap [(TileY << 7) + TileX]]
                   .SpriteData [(TPixY << 4) + TPixX];
  }
}

void main () {
  //Temporary x and y variables
  unsigned int   tX, tY;
  //What x and y coordinates should be displayed at the top left of the screen
  int          tlX, tlY;
  //The direction the screen should bounce around the map.
  int        xdir, ydir;

  //Switch to graphics mode
  VGAScreen ();

  #ifdef __DJGPP
    //Get a pointer to the screen
    __djgpp_nearptr_enable ();
    Screen = (char *) (__djgpp_conventional_base + 0xA0000);
  #else
    //Get a pointer to the screen and assign memory to the buffer
    Screen = (unsigned char *)MK_FP (0xA000, 0000);
    Buffer = (unsigned char far *) farmalloc(64000);
  #endif

  //Clear the buffer
  ClearBuffer ();

  //Set up a sprite consisting of a blue block
  for (tX = 0; tX < 16; tX++) {
    for (tY = 0; tY < 16; tY++) {
      PutPixel (tX, tY, 1);
    }
  }
  MakeSprite (&Tiles [0], 0, 0, 15, 15);

  //Set up a sprite consisting of a red rectangle
  for (tX = 0; tX < 16; tX++) {
    PutPixel (16 + tX,  0, 4);
    PutPixel (16 + tX, 15, 4);
    PutPixel (16 +  0, tX, 4);
    PutPixel (16 + 15, tX, 4);
  }
  MakeSprite (&Tiles [1], 16, 0, 16 + 15, 15);

  //Set up a tiled background of 128x128 tiles
  for (tY = 0; tY < 128; tY++) for (tX = 0; tX < 128; tX++) {
    //This generates a checkerboard...
    TileMap [(tY << 7) + tX] = ((tY & 1) + tX) & 1;
  }

  //Start at top left corner, moving down-right
  tlX = tlY = 0;
  xdir = ydir = 2;

  do {
    //Draw a tiled background, with coordinates tlX and tlY.
    DrawTiledBackground (tlX, tlY);

    //Display the buffer - if you leave this out, you won't see a thing
    DisplayBuffer ();

    //Update coordinates to bounce around the screen
    tlX += xdir;
    tlY += ydir;
    if (tlX < 0) { xdir = 1; tlX = 0; }
    if (tlY < 0) { ydir = 1; tlY = 0; }
    if (tlX > (128 * 16 - 320)) { xdir = -xdir; tlX = (128 * 16 - 320); }
    if (tlY > (128 * 16 - 200)) { ydir = -ydir; tlY = (128 * 16 - 200); }
  } while (!kbhit());
  getch ();
  TextScreen ();
}